import os


def make_folder(folder, **kwargs):
    """Utility to make folders

    Args:
        folder (string): name of folder

    Returns:
        string: path to folder
    """
    # Makes folder
    os.makedirs(folder, exist_ok=True)

    return folder



import numpy as np
import torch
import torch.nn as nn

tanh = nn.Tanh()
selu = nn.SELU()
sigmoid = nn.Sigmoid()

def default_nl_function(t, x, y, z):
  
  # returns a function from variables
  return tanh(torch.tensor(20*(t - 2*(x-.5)))) + selu(torch.tensor((t-2*(y-0.5)))) + sigmoid(torch.tensor(-20*(t-(z-0.5))))

def generate_data(values, function=default_nl_function, length=25, range_=[-1, 1]):
  """_summary_

  Args:
      values (array): Input values to use
      function (obj, optional): Function to use for generation. Defaults to default_nl_function.
      length (int, optional): length of the vector to generate. Defaults to 25.
      range_ (list, optional): range of spectra where you generate data. Defaults to [-1, 1].

  Returns:
      array: computed spectra
  """
  # build x vector
  x = np.linspace(range_[0], range_[1], length)

  data = np.zeros((values.shape[0], length))

  for i in range(values.shape[0]):
      data[i, :] = function(x, values[i, 0], values[i, 1], values[i, 2])

  return data


import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
import os
import torch


def layout_fig(graph, mod=None):

    # """Utility function that helps lay out many figures

    # Args:
    #     graph (int): number of graphs
    #     mod (int, optional): value that assists in determining the number of rows and columns. Defaults to None.

    # Returns:
    #     tuple: figure and axis
    # """

    # Sets the layout of graphs in matplotlib in a pretty way based on the number of plots
    if mod is None:
        # Select the number of columns to have in the graph
        if graph < 3:
            mod = 2
        elif graph < 5:
            mod = 3
        elif graph < 10:
            mod = 4
        elif graph < 17:
            mod = 5
        elif graph < 26:
            mod = 6
        elif graph < 37:
            mod = 7

    # builds the figure based on the number of graphs and a selected number of columns
    fig, axes = plt.subplots(
        graph // mod + (graph % mod > 0),
        mod,
        figsize=(3 * mod, 3 * (graph // mod + (graph % mod > 0))),
    )

    # deletes extra unneeded axes
    axes = axes.reshape(-1)
    for i in range(axes.shape[0]):
        if i + 1 > graph:
            fig.delaxes(axes[i])

    return fig, axes


def embedding_maps(data, image, colorbar_shown=True, c_lim=None, mod=None, title=None):
    """function that generates the embedding maps

    Args:
        data (array): embedding maps to plot
        image (array): raw image used for the sizing of the image
        colorbar_shown (bool, optional): selects if colorbars are shown. Defaults to True.
        c_lim (array, optional): sets the range for the color limits. Defaults to None.
        mod (int, optional): used to change the layout (rows and columns). Defaults to None.
        title (string, optional): Adds title to the image . Defaults to None.
    """
    fig, ax = layout_fig(data.shape[1], mod)

    for i, ax in enumerate(ax):
        if i < data.shape[1]:
            im = ax.imshow(data[:, i].reshape(image.shape[0], image.shape[1]))
            ax.set_xticklabels("")
            ax.set_yticklabels("")

            # adds the colorbar
            if colorbar_shown is True:
                divider = make_axes_locatable(ax)
                cax = divider.append_axes("right", size="10%", pad=0.05)
                cbar = plt.colorbar(im, cax=cax, format="%.1e")

                # Sets the scales
                if c_lim is not None:
                    im.set_clim(c_lim)

    if title is not None:
        # Adds title to the figure
        fig.suptitle(title, fontsize=16, y=1, horizontalalignment="center")

    fig.tight_layout()


def imagemap(ax, data, colorbars=True, clim=None):
    """pretty way to plot image maps with standard formats

    Args:
        ax (ax): axes to write to
        data (array): data to write
        colorbars (bool, optional): selects if you want to show a colorbar. Defaults to True.
        clim (array, optional): manually sets the range of the colorbars. Defaults to None.
    """

    if data.ndim == 1:
        data = data.reshape(
            np.sqrt(data.shape[0]).astype(int), np.sqrt(data.shape[0]).astype(int)
        )

    cmap = plt.get_cmap("viridis")

    if clim is None:
        im = ax.imshow(data, cmap=cmap)
    else:
        im = ax.imshow(data, clim=clim, cmap=cmap)

    ax.set_yticklabels("")
    ax.set_xticklabels("")

    if colorbars:
        # adds the colorbar
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="10%", pad=0.05)
        cbar = plt.colorbar(im, cax=cax, format="%.1e")


def find_nearest(array, value, averaging_number):
    """computes the average of some n nearest neighbors

    Args:
        array (array): input array
        value (float): value to find closest to
        averaging_number (int): number of data points to use in averaging

    Returns:
        list: list of indexes of the nearest neighbors
    """
    idx = (np.abs(array - value)).argsort()[0:averaging_number]
    return idx


def latent_generator(
    model,
    embeddings,
    image,
    number,
    average_number,
    indx=None,
    ranges=None,
    x_values=None,
    y_scale=[-2.2, 4],
    device="cuda",
):
    """Plots the generator results

    Args:
        model (PyTorch object): neural network model
        embeddings (float, array): the input embedding (or output from the encoder)
        image (array): Original image, this is used to extract the size of the embedding
        number (int): number of divisions to plot
        average_number (int): number of samples to average in the generation process
        indx (list, optional): embedding indexes to use. Defaults to None.
        ranges (float, array, optional): set the ranges for the embeddings. Defaults to None.
        x_values (array, optional): allows addition of x_values. Defaults to None.
        y_scale (list, optional): Scale of the y-axis. Defaults to [-2.2, 4].
        device (str, optional): the device where the data will be processed. Defaults to 'cuda'.
    """

    # sets the colormap
    cmap = plt.cm.viridis

    if indx is None:
        embedding_small = embeddings.squeeze()
    else:
        embedding_small = embeddings[:, indx].squeeze()

    # creates the figures and axes in a pretty way
    fig, ax = layout_fig(embedding_small.shape[1] * 2, mod=3)

    # plots all of the embedding maps
    for i in range(embedding_small.shape[1]):
        im = imagemap(
            ax[i], embedding_small[:, i].reshape(image.shape[0], image.shape[1])
        )

    # loops around the number of example loops
    for i in range(number):

        # loops around the number of embeddings from the range file
        for j in range(embedding_small.shape[1]):

            if ranges is None:
                value = np.linspace(
                    np.min(embedding_small[:, j]), np.max(embedding_small[:, j]), number
                )
            else:
                # sets the linear spaced values
                value = np.linspace(0, ranges[j], number)

            idx = find_nearest(embedding_small[:, j], value[i], average_number)
            gen_value = np.mean(embeddings[idx], axis=0)
            gen_value[j] = value[i]

            # computes the generated results
            gen_value_1 = torch.from_numpy(np.atleast_2d(gen_value)).to(device)
            generated = model(gen_value_1)
            generated = generated.to("cpu")
            generated = generated.detach().numpy().squeeze()

            # plots and formats the graphs
            if x_values is None:
                ax[j + embedding_small.shape[1]].plot(
                    generated, color=cmap((i + 1) / number)
                )
            else:
                ax[j + embedding_small.shape[1]].plot(
                    x_values, generated, color=cmap((i + 1) / number)
                )

            ax[j + embedding_small.shape[1]].set_ylim(y_scale)
            # ax[j + embedding_small.shape[1]].set_yticklabels('')
            plt.tight_layout(pad=1)



from setuptools import setup, find_packages
import os

setup(
    name='m3_learning',
    version='0.0.2',
    packages=find_packages(where="src"),
    url='https://github.com/jagar2/m3_learning.git',
    # install_requires=requirements,
    license=' BSD-3-Clause',
    author='Joshua C. Agar',
    author_email='jca92@drexel.edu',
    description='Tutorials, Projects, and datasets from the M3-learning research group',
    classifiers = [
                  "Programming Language :: Python :: 3",
                  "License :: OSI Approved :: BSD License",
                  "Operating System :: OS Independent",
              ],
              package_dir = {"": "src"},
                            python_requires = ">=3.6",)


import torch.nn as nn
import torch.nn.functional as F
import torch
from tqdm import tqdm
import numpy as np


class Regularization(nn.Module):
    def __init__(self, model, weight_decay, p=2, device="cuda"):
        """init function

        Args:
            model (PyTorch model): neural network model
            weight_decay (float): value for the weight decay
            p (int, optional): l1 regularization. Defaults to 2.
            device (str, optional): the device where the model is located. Defaults to 'cuda'.
        """

        super(Regularization, self).__init__()
        if weight_decay < 0:
            print("param weight_decay can not <0")
            exit(0)
        self.model = model
        self.weight_decay = weight_decay
        self.p = p
        self.weight_list = self.get_weight(model)
        self.device = device

    def to(self, device):
        """function that sets the device

        Args:
            device (string): PyTorch device

        Returns:
            obj: self
        """
        super().to(device)
        return self

    def forward(self, model):
        """Conducts the forward pass of the model

        Args:
            model (PyTorch model): model

        Returns:
            float: computed regularization loss
        """
        self.weight_list = self.get_weight(model)
        reg_loss = self.regularization_loss(
            self.weight_list, self.weight_decay, p=self.p
        )
        return reg_loss

    def get_weight(self, model):
        """_summary_

        Args:
            model (PyTorch model): model

        Returns:
            list: list of weights
        """
        weight_list = []
        for name, param in model.named_parameters():
            if "dec" in name and "weight" in name:
                weight = (name, param)
                weight_list.append(weight)
        return weight_list

    def regularization_loss(self, weight_list, weight_decay, p):
        """Calculates the regularization loss

        Args:
            weight_list (list): list of weights
            weight_decay (float): Sets how the regularization is decayed
            p (float): sets the norm that is used

        Returns:
            _type_: _description_
        """
        reg_loss = 0
        for name, w in weight_list:
            l2_reg = torch.norm(w, p=p)
            reg_loss = reg_loss + l2_reg

        reg_loss = weight_decay * reg_loss
        return reg_loss

    def weight_info(self, weight_list):
        """List of weights in layers to regularize

        Args:
            weight_list (list): list of weights
        """
        print("---------------regularization weight---------------")
        for name, w in weight_list:
            print(name)


def loss_function(
    model,
    encoder,
    decoder,
    train_iterator,
    optimizer,
    coef=0,
    coef1=0,
    ln_parm=1,
    beta=None,
    mse=True,
    device="cuda",
):

    """Loss function

    Args:
        model (PyTorch model): model
        encoder (PyTorch model): encoder of the mode
        decoder (PyTorch model): decoder of the model
        train_iterator (iter): iterator used from straining
        optimizer (obj): optimization methods used
        coef (int, optional): used to set the lambda value of the regularization. Defaults to 0.
        coef1 (int, optional): not implemented. Defaults to 0.
        ln_parm (int, optional): norm value. Defaults to 1.
        beta (float, optional): beta for variational autoencoder. Defaults to None.
        mse (bool, optional): selects to use the MSE loss. Defaults to True.
        device (str, optional): selects the device to use. Defaults to "cuda".

    Returns:
        _type_: _description_
    """

    # regularization coefficients
    weight_decay = coef
    weight_decay_1 = coef1

    # set the training mode
    model.train()

    # loss of the epoch
    train_loss = 0
    #    for i, x in enumerate(train_iterator):
    for x in tqdm(train_iterator, leave=True, total=len(train_iterator)):

        # calculates regularization on the entire model
        reg_loss_2 = Regularization(model, weight_decay_1, p=2).to(device)

        x = x.to(device, dtype=torch.float)

        # update the gradients to zero
        optimizer.zero_grad()

        if beta is None:
            embedding = encoder(x)

        else:

            # forward pass
            #        predicted_x = model(x)
            embedding, sd, mn = encoder(x)

        if weight_decay > 0:
            reg_loss_1 = weight_decay * torch.norm(embedding, ln_parm).to(device)
        else:
            reg_loss_1 = 0.0

        predicted_x = decoder(embedding)

        if mse:
            # reconstruction loss
            loss = F.mse_loss(x, predicted_x, reduction="mean")
        else:
            # reconstruction loss
            loss = F.mse_loss(x, predicted_x, reduction="mean")

            loss = loss + reg_loss_2(model) + reg_loss_1

        # beta VAE
        if beta is not None:
            vae_loss = (
                beta * 0.5 * torch.sum(torch.exp(sd) + (mn) ** 2 - 1.0 - sd).to(device)
            )
            vae_loss /= sd.shape[0] * sd.shape[1]
        else:
            vae_loss = 0

        loss = loss + vae_loss

        # backward pass
        train_loss += loss.item()

        loss.backward()
        # update the weights
        optimizer.step()

    return train_loss


def Train(
    model,
    encoder,
    decoder,
    train_iterator,
    optimizer,
    epochs,
    coef=0,
    coef_1=0,
    ln_parm=1,
    beta=None,
    mse=True,
    device="cuda",
    save_weight = False,
):
    """Function that trains the model

    Args:
        model (Pytorch model): autoencoder model
        encoder (PyTorch model): encoder of the mode
        decoder (PyTorch model): decoder of the model
        train_iterator (iter): iterator used from straining
        optimizer (obj): optimization methods used
        epochs (int): number of epochs
        coef (int, optional): used to set the lambda value of the regularization. Defaults to 0.
        coef1 (int, optional): not implemented. Defaults to 0.
        ln_parm (int, optional): norm value. Defaults to 1.
        beta (float, optional): beta for variational autoencoder. Defaults to None.
        mse (bool, optional): selects to use the MSE loss. Defaults to True.
        device (str, optional): selects the device to use. Defaults to "cuda".

    """
    N_EPOCHS = epochs
    best_train_loss = float("inf")

    for epoch in range(N_EPOCHS):

        train = loss_function(
            model,
            encoder,
            decoder,
            train_iterator,
            optimizer,
            coef,
            coef_1,
            ln_parm,
            beta,
            mse,
            device,
        )

        train_loss = train
        train_loss /= len(train_iterator)
        print(f"Epoch {epoch}, Train Loss: {train_loss:.4f}")
        print(".............................")

        if best_train_loss > train_loss:
            best_train_loss = train_loss

            checkpoint = {
                "net": model.state_dict(),
                "optimizer": optimizer.state_dict(),
                "epoch": epoch,
                "encoder": encoder.state_dict(),
                "decoder": decoder.state_dict(),
            }
            if save_weight:
                if epoch >= 0:
                    torch.save(
                        checkpoint, f"./test__Train Loss:{train_loss:.4f}-{epoch}.pkl"
                    )


def transform_nn(data, encoder, decoder, device = 'cuda'):
    """Extracts the inference from the autoencoder

    Args:
        data (array, float): input data
        encoder (PyTorch model): encoder block
        decoder (PyTorch model): decoder block
        device (str, optional): selects the device to use. Defaults to "cuda".

    Returns:
        array, float: encoder results, decoder results
    """
    try:
        encoded_spectra = encoder(
            torch.tensor(np.atleast_3d(data), dtype=torch.float32).to(device)
        )
    except:
        pass

    try:
        encoded_spectra = encoder(torch.tensor(data, dtype=torch.float32).to(device))
    except:
        pass

    decoded_spectra = decoder(encoded_spectra)

    encoded_spectra = encoded_spectra.to("cpu")
    encoded_spectra = encoded_spectra.detach().numpy()
    decoded_spectra = decoded_spectra.to("cpu")
    decoded_spectra = decoded_spectra.detach().numpy()
    return encoded_spectra, decoded_spectra
